package uk.co.keang.swingaddons;

import java.awt.Component;
import java.awt.Container;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.ContainerEvent;
import java.awt.event.ContainerListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JViewport;

/**
 * Listener to keep track of the components as they are added and removed from
 * this wizard page.
 * 
 * @author rogue
 */
public class FocusAutoScroller implements ContainerListener
{
private final FocusListener focusListener;
private final JViewport viewport;
private final ViewComponent compToView;
private BiasH alignH;
private BiasV alignV;
private Insets insets = new Insets(0, 0, 0, 0);
private boolean scrollToParent;

public enum ViewComponent
{
CHILD {
        @Override
        Component getComponentToView(Component c)
            {
            // this is the child so just return it
            return c;
            }
    },
PARENT {
        @Override
        Component getComponentToView(Component c)
            {
            // get the parent
            return c.getParent();
            }
    };

abstract Component getComponentToView(Component c) ;
}

public enum BiasH
{
LEFT
    {
    @Override
    void adjustViewport(Rectangle cRect, Rectangle vRect, Insets insets)
        {
        // if there's room to use insets do so
        if ( vRect.width - insets.left >= cRect.width )
            cRect.x -= insets.left;
        else if ( vRect.width >= cRect.width )
            cRect.x -= vRect.width - cRect.width;

        cRect.width = vRect.width;
        }
    },
CENTER
    {
    @Override
    void adjustViewport(Rectangle cRect, Rectangle vRect, Insets insets)
        {
        int vMidx = vRect.width / 2;
        cRect.x -= vMidx - (cRect.width / 2);

        cRect.width = vRect.width;
        }
    },
RIGHT
    {
    @Override
    void adjustViewport(Rectangle cRect, Rectangle vRect, Insets insets)
        {
        int vW;

        // if there's room to use insets do so
        if ( vRect.width - insets.right >= cRect.width )
            vW = vRect.width - insets.right;
        else if ( vRect.width >= cRect.width )
            vW = cRect.width;
        else
            vW = vRect.width;

        cRect.x -= (vW - cRect.width);
        if ( cRect.x < 0 )
            cRect.x = 0;

        cRect.width = vRect.width;
        }
    };

abstract void adjustViewport(Rectangle cRect, Rectangle vRect, Insets insets) ;
}

public enum BiasV
{
TOP
    {
    @Override
    void adjustViewport(Rectangle cRect, Rectangle vRect, Insets insets)
        {
        // if there's room to use insets do so
        if ( vRect.height - insets.top >= cRect.height )
            cRect.y -= insets.top;
        else if ( vRect.height >= cRect.height )
            cRect.y -= vRect.height - cRect.height;

        cRect.height = vRect.height;
        }
    },
CENTER
    {
    @Override
    void adjustViewport(Rectangle cRect, Rectangle vRect, Insets insets)
        {
        int vMidy = vRect.height / 2;
        cRect.y -= vMidy - (cRect.height / 2);

        cRect.height = vRect.height;
        }
    },
BOTTOM
    {
    @Override
    void adjustViewport(Rectangle cRect, Rectangle vRect, Insets insets)
        {
        int vH;

        // if there's room to use insets do so
        if ( vRect.height - insets.bottom >= cRect.height )
            vH = vRect.height - insets.bottom;
        else if ( vRect.height >= cRect.height )
            vH = cRect.height;
        else
            vH = vRect.height;

        cRect.y -= (vH - cRect.height);
        if ( cRect.y < 0 )
            cRect.y = 0;

        cRect.height = vRect.height;
        }
    };
abstract void adjustViewport(Rectangle cRect, Rectangle vRect, Insets insets) ;
}

/**
 * Creates a scroller for this scroll pane based on each child component
 * with horizontal bias of right and vertical bias of center
 *  
 * @param scrollPane the scrollpane to scroll.
 */
public FocusAutoScroller(JScrollPane scrollPane)
    {
    this(scrollPane, ViewComponent.CHILD, BiasH.RIGHT, BiasV.CENTER);
    }

public FocusAutoScroller(JScrollPane scrollPane, ViewComponent comp, BiasH h, BiasV v)
    {
    viewport = scrollPane.getViewport();
    compToView = comp;
    alignH = h;
    alignV = v;
    
    // create the focus listener
    focusListener = new FocusListener()
        {
        private Component prevComp;
        
        @Override
        public void focusLost(FocusEvent e)
            {
            // do nothing
            }
        
        @Override
        public void focusGained(FocusEvent e)
            {
            // make sure component is visible
            Component src = compToView.getComponentToView((Component)e.getSource());

            // Ignore this request if we've already handled this component.
            // This prevents scrolling occurring if the scroll area has been manually changed
            // and the app loses and regains focus. 
            if ( src == prevComp )
                return;
            
            prevComp = src;
            
            if ( scrollToParent )
                src = src.getParent();
            
            // get the size and location of the component
            Rectangle rect = src.getBounds();
            
            // adjust the location so it is relative to the root panel
            Container parent = src.getParent();
            Component root = viewport.getView();
            while ( parent != root )
                {
                // offset the location so the value is relative to the root scroll pane
                Rectangle pRect = parent.getBounds();
                
                rect.x += pRect.x;
                rect.y += pRect.y;
                
                parent = parent.getParent();
                }

            // adjust the size and location for the specified biases
            alignH.adjustViewport(rect, viewport.getViewRect(), insets);
            alignV.adjustViewport(rect, viewport.getViewRect(), insets);
            
            // scroll so the rectangle is visible
            ((JComponent)root).scrollRectToVisible(rect);
            }
        };
        
    // add listeners to existing components    
    Component root = viewport.getView();
    if ( root != null )
        addListener(root);
    }

/**
 * Gets the insets applied when auto scrolling
 * 
 * @return the insets
 */
public Insets getScrollInsets()
    {
    return insets;
    }

/**
 * Sets the insets to apply when auto scrolling
 * 
 * @param ins the insets
 */
public void setScrollInsets(Insets ins)
    {
    insets = ins;
    }

/**
 * Sets the insets to apply when auto scrolling
 * 
 * @param top
 * @param left
 * @param bottom
 * @param right
 */
public void setScrollInsets(int top, int left, int bottom, int right)
    {
    insets = new Insets(top, left, bottom, right);
    }

/**
 * Sets whether scrolling should be done based on the focused component
 * or it's parent. If components are grouped in sub panels and you
 * want the whole sub panel to be scrolled to when one of it's components
 * has focus, set this to true.
 * <p>
 * By default this is false ie scrolling is based on the component
 * 
 * @param state
 */
public void setScrollToParent(boolean state)
    {
    scrollToParent = state;
    }

/*
 * (non-Javadoc)
 * 
 * @see
 * java.awt.event.ContainerListener#componentAdded(java.awt.event.ContainerEvent
 * )
 */
@Override
public void componentAdded(ContainerEvent e)
    {
    Component newComp = e.getChild();

    if ( newComp instanceof JComponent )
        {
        addListener(newComp);
        }
    }

/**
 * Adds a listener to the component.
 * Recursively calls embedded panels and tabbed panes 
 * 
 * @param newComp
 */
private void addListener(Component newComp)
    {
    if ( newComp instanceof JScrollPane )
        {
        // recurse into the view component
        addListener(((JScrollPane)newComp).getViewport().getView());
        return;
        }
    else if ( newComp instanceof JPanel || newComp instanceof JTabbedPane )
        {
        // recurse:
        Component[] children = ((Container)newComp).getComponents();
        for ( Component c : children )
            {
            addListener(c);
            }
        
        ((Container)newComp).addContainerListener(this);
        return;
        }

    newComp.addFocusListener(focusListener);
    }

/*
 * (non-Javadoc)
 * 
 * @see
 * java.awt.event.ContainerListener#componentRemoved(java.awt.event.ContainerEvent
 * )
 */
@Override
public void componentRemoved(ContainerEvent e)
    {
    componentRemoved(e.getChild());
    }

/**
 * Recursively remove the focus listener from the component
 * 
 * @param comp the component being removed
 */
private void componentRemoved(Component comp)
    {
    if ( comp instanceof Container )
        {
        Component[] comps = ((Container)comp).getComponents();
        for ( Component c : comps )
            {
            componentRemoved(c);
            }
        }
    
    comp.removeFocusListener(focusListener);
    }

/**
 * Sets the horizontal bias
 * 
 * @param align the bias
 */
public void setHorizontalBias(BiasH align)
    {
    alignH = align;
    }

/**
 * Gets the horizontal bias
 * 
 * @return align the bias
 */
public BiasH getHorizontalBias()
    {
    return alignH;
    }

/**
 * Sets the vertical bias
 * 
 * @param the bias
 */
public void setVerticalBias(BiasV align)
    {
    alignV = align;
    }

/**
 * Gets the vertical bias
 * 
 * @return the bias
 */
public BiasV getVerticalBias()
    {
    return alignV;
    }

}

