/**
 * Copyright (c) 2008 A.G.Docherty. All Rights Reserved.
 */
package uk.co.keang.swingaddons;
 
import java.awt.Component;
import java.awt.EventQueue;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.swing.JComponent;

/**
 * Recognises gestures and notifies listeners. 
 * Compound gestures are also recognised, although by default this feature is disabled.
 * 
 * Two notification mechanisms are supported. 
 * 1. Basic Listener registration and notification.
 * 2. A mapping of Gesture types to an ID and a mapping from the ID to an Action Object. 
 * 
 * @author A.G.Docherty
 * @date 18 Dec 2008
 *
 */
public class Gesture implements MouseMotionListener, MouseListener
{
private static final int GESTURE_ID = ActionEvent.ACTION_LAST + 1;

private final List<IGestureListener> listeners = new ArrayList<IGestureListener>();
private final List<Direction> gestures = new ArrayList<Direction>();
private JComponent src;
private Point start;
private int minMovement = 40;
private int minMovementSq = minMovement * minMovement;
private int boundary = (int)(minMovement/2.5f);
private boolean compoundEnabled = false;

private Direction currentGesture;

private boolean enabled;

private final Map<String, ActionListener> actionMap = new HashMap<String, ActionListener>(30);
private final Map<List<Direction>, String> gestureMap = new HashMap<List<Direction>, String>(30);

public static enum Direction
{
UP("\u2191"),
DOWN("\u2193"),
LEFT("\u2190"),
RIGHT("\u2192"),
TOP_LEFT_TO_BOTTOM_RIGHT("\u2198"),
TOP_RIGHT_TO_BOTTOM_LEFT("\u2199"),
BOTTOM_LEFT_TO_TOP_RIGHT("\u2197"),
BOTTOM_RIGHT_TO_TOP_LEFT("\u2196");

private final String arrow;

Direction(String symbol)
    {
    arrow = symbol;
    }

/**
 * Get the arrow symbol associated with this gesture
 * 
 * @return an arrow symbol
 */
public String getSymbol() { return arrow; }

/**
 * Gets the GESTURE associated with this arrow symbol
 * 
 * @param symbol the symbol to search for
 * @return the GESTURE for this symbol or null if there isn't one
 * @throws IllegalArgumentException if there is no GESTURE for this symbol
 */
public static Direction getGesture(String symbol)
    {
    for ( Direction g : Direction.values() )
        {
        if ( g.arrow.equals(symbol) )
            return g;
        }
    
    throw new IllegalArgumentException("Symbol "+symbol+" is not a valid gesture");
    }

/**
 * Tests to see if there is a GESTURE associated with this arrow symbol
 * 
 * @param symbol the symbol to search for
 * @return true if there is a GESTURE for this symbol 
 */
public static boolean hasGesture(String symbol)
    {
    for ( Direction g : Direction.values() )
        {
        if ( g.arrow.equals(symbol) )
            return true;
        }
    
    return false;
    }
}

/**
 * Adds a component as a source of gestures
 * 
 * @param c - the source component
 */
public void addComponent(Component c)
    {
    c.addMouseListener(this);
    c.addMouseMotionListener(this);
    }

/**
 * Remove a component as a source of gestures
 * 
 * @param c - the source component
 */
public void removeComponent(Component c)
    {
    c.removeMouseListener(this);
    c.removeMouseMotionListener(this);
    }

/**
 * Adds a listener of Gestures
 * 
 * @param listener - the listener to add
 */
public void addGestureListener(IGestureListener listener)
    {
    listeners.add(listener);
    }

/**
 * Removes a listener of Gestures
 * 
 * @param listener - the listener to add
 */
public void removeGestureListener(IGestureListener listener)
    {
    listeners.remove(listener);
    }

/**
 * Notifies all listeners a gesture has been recognised
 * 
 * @param g - the recognised gesture pattern
 * @param s - the source component
 */
private void fireGesture(List<Direction> g, JComponent s)
    {
    final Direction[] gArray = gestures.toArray(new Direction[gestures.size()]);

    for ( IGestureListener listener : listeners )
        {
        listener.gestureRecognised(gArray, s);
        }
    }

/**
 * Notifies all action object that a gesture has been recognised.
 * 
 * @param g - the recognised gesture pattern
 * @param s - the source component
 */
private void fireGestureAction(List<Direction> g, JComponent s)
    {
    String id = gestureMap.get(g);
    
    // if there is no mapping for this gesture return now
    if ( id == null )
        return;
    
    ActionListener a = actionMap.get(id);
    
    // if there is no mapping for this gesture ID return now
    if ( a == null )
        return;
    
    a.actionPerformed(new ActionEvent(s, GESTURE_ID, id));
    }

@Override
public void mousePressed(MouseEvent e)
    {
    if ( !enabled )
        return;
    
    src = (JComponent)e.getSource();
    start = e.getPoint();
    
    synchronized ( gestures )
        {
        currentGesture = null;
        gestures.clear();
        }
    }

@Override
public void mouseReleased(MouseEvent e)
    {
    if ( !enabled )
        return;
    
    processMovement(e);
    
    // if there's been a gesture then consume this event
    if ( gestures.size() > 0 )
        {
        e.consume();

        EventQueue.invokeLater(new Runnable()
            {
            @Override
            public void run()
                {
                fireGestureAction(gestures, src);
                fireGesture(gestures, src);
                }
            });
        }
    }

@Override
public void mouseDragged(MouseEvent event)
    {
    if ( !enabled )
        return;
    
    processMovement(event);
    }

/**
 * Processes the mouse movement into a gesture
 * 
 * @param event - the mouse event
 */
private void processMovement(MouseEvent event)
    {
    // if compound gestures are disabled and we already have a gesture then return now
    if ( !compoundEnabled && currentGesture != null )
    	return;
    
    Point point = event.getPoint();
    
    int deltaX = start.x - point.x;
    int deltaY = start.y - point.y;
    int absDeltaX = Math.abs(deltaX);
    int absDeltaY = Math.abs(deltaY);

    // if we haven't moved far enough for this to be a gesture
    if ( (absDeltaX * absDeltaX) + (absDeltaY * absDeltaY) < minMovementSq )
    	{
        // if there isn't a gesture then exit now
        if ( currentGesture == null )
            return;

        // check to see if this movement is a continuation of the current gesture or
        // if it is a different gesture eg UP followed by DOWN, or DOWN followed by RIGHT
        
        // if we've moved as far as the boundary
        if ( absDeltaX > boundary || absDeltaY > boundary )
            {
            // if the motion is basically in the same direction as the current gesture 
            // then consume these events
            switch ( currentGesture )
                {
                case UP:
                    if ( deltaY >= boundary && absDeltaX < boundary/2)
                        start = point;
                    break;

                case DOWN:
                    if ( absDeltaY >= boundary && absDeltaX < boundary/2 )
                        start = point;
                    break;

                case LEFT:
                    if ( deltaX >= boundary && absDeltaY < boundary/2 )
                        start = point;
                    break;

                case RIGHT:
                    if ( absDeltaX >= boundary && absDeltaY < boundary/2 )
                        start = point;
                    break;

                case TOP_LEFT_TO_BOTTOM_RIGHT:
                    if ( deltaY < 0 && deltaX < 0 )
                        start = point;
                    break;

                case TOP_RIGHT_TO_BOTTOM_LEFT:
                    if ( deltaY < 0 && deltaX > 0 )
                        start = point;
                    break;

                case BOTTOM_LEFT_TO_TOP_RIGHT:
                    if ( deltaY > 0 && deltaX < 0 )
                        start = point;
                    break;

                case BOTTOM_RIGHT_TO_TOP_LEFT:
                    if ( deltaY > 0 && deltaX > 0 )
                        start = point;
                    break;

                default:
                	throw new IllegalArgumentException("The gesture "+currentGesture+" is handled");
                }
            }
        
        return;
        }
    
    if ( absDeltaX >= 0 && absDeltaX < boundary )
        {
        // a vertical movement
        if ( deltaY < 0 )
            addGesture(Direction.DOWN);
        else
            addGesture(Direction.UP);
        }
    else if ( absDeltaY >= 0 && absDeltaY < boundary )
        {
        // a horizontal movement
        if ( deltaX > 0 )
            addGesture(Direction.LEFT);
        else
            addGesture(Direction.RIGHT);
        }
    else if ( deltaY < 0 )
        {
        // a diagonal movement from top down
        if ( deltaX < 0 )
            addGesture(Direction.TOP_LEFT_TO_BOTTOM_RIGHT);
        else
            addGesture(Direction.TOP_RIGHT_TO_BOTTOM_LEFT);
        }
    else
        {
        // a diagonal movement from bottom up
        if ( deltaX < 0 )
            addGesture(Direction.BOTTOM_LEFT_TO_TOP_RIGHT);
        else
            addGesture(Direction.BOTTOM_RIGHT_TO_TOP_LEFT);
        }
    
    start = point;
    }

/**
 * Adds a gesture to the list of gestures
 * 
 * @param g - the gesture to add
 */
private void addGesture(Direction g)
    {
    synchronized ( gestures )
        {
        // if this is the same gesture then ignore it
        int size = gestures.size();
        if ( size > 0 && gestures.get(size-1) == g )
            return;
            
        currentGesture = g;
        gestures.add(g);
        }
    }

@Override
public void mouseMoved(MouseEvent e)
    {
    // not required
    }

@Override
public void mouseClicked(MouseEvent e)
    {
    // not required
    }

@Override
public void mouseEntered(MouseEvent e)
    {
    // not required
    }

@Override
public void mouseExited(MouseEvent e)
    {
    // not required
    }

/**
 * Gets the minimum movement required to generate a gesture
 *
 * @return the minimum movement in pixels
 */
public int getMinMovement()
    {
    return minMovement;
    }

/**
 * Sets the minimum movement required to generate a gesture
 *
 * @param min - the minMovement in pixels
 */
public void setMinMovement(int min)
    {
    minMovement = min;
    minMovementSq = minMovement * minMovement;
    boundary = (int)(minMovement/2.5f);
    }

/**
 * Sets the enabled state of the gesture handler
 * 
 * @param state - true to enable gesture recognition
 */
public void setEnabled(boolean state)
    {
    synchronized ( gestures )
        {
        enabled = state;
        
        // if gestures are disabled clear the gesture info now 
        if ( !enabled )
            {
            currentGesture = null;
            gestures.clear();
            }
        }
    }

/**
 * Adds an action with this identifier
 * 
 * @param name - the id
 * @param act - the action
 */
public void addAction(String name, ActionListener act)
    {
    if ( name == null )
        throw new NullPointerException("Action name cannot be null");

    actionMap.put(name, act);
    }

/**
 * Removes an action with this identifier
 * 
 * @param name - the id
 */
public void removeAction(String name)
    {
    actionMap.remove(name);
    }

/**
 * Removes all gesture mappings
 */
public void removeAllActionMapping()
    {
    actionMap.clear();
    }

/**
 * Adds a gesture mapping
 * 
 * @param gesture - the gesture
 * @param name - the action id
 */
public void addGestureActionMapping(Direction gesture, String name)
    {
    List<Direction> list = new ArrayList<Direction>();
    list.add(gesture);
    
    addGestureActionMapping(list, name);
    }

/**
 * Adds a gesture mapping
 * 
 * @param gesture - the compound gesture
 * @param name - the action id
 */
public void addGestureActionMapping(List<Direction> gesture, String name)
    {
    if ( name == null )
        throw new NullPointerException("Action name cannot be null");
    
    if ( gesture == null )
        throw new NullPointerException("Gesture cannot be null");
    
    if ( gesture.size() == 0 )
        throw new IllegalArgumentException("There must be at least one gesture specified");
    
    gestureMap.put(gesture, name);
    }

/**
 * Removes a gesture mapping
 * 
 * @param g - the gesture
 */
public void removeGestureActionMapping(List<Direction> g)
    {
    gestureMap.remove(g);
    }

/**
 * Removes a gesture mapping
 * 
 * @param name - the action id
 */
public void removeGestureActionMapping(String name)
    {
    Set<Entry<List<Direction>, String>> m = gestureMap.entrySet();
    
    for ( Entry<List<Direction>, String> entry : m )
        {
        if ( entry.getValue().equals(name) )
            {
            List<Direction> key = entry.getKey();
            gestureMap.remove(key);
            return;
            }
        }
    }

/**
 * Removes all gesture mappings
 */
public void removeAllGestureMapping()
    {
    gestureMap.clear();
    }

/**
 * Gets the compound gesture enabled state
 *
 * @return true if compound gestures will be detected
 */
public boolean isCompoundEnabled()
	{
	return compoundEnabled;
	}

/**
 * Sets the compound gestures enabled state. 
 * If compound gestures are not enabled only the first gesture is detected and all
 * subsequent movements are ignored.
 *
 * @param state true to allow compound gestures to be detected
 */
public void setCompoundEnabled(boolean state)
	{
	compoundEnabled = state;
	}
}
