/** 
 * Copyright (c) 2010 Keang Ltd.
 * 
 * This file is part of the Swing Add On's package.
 *
 * The Swing Add On's package is free software: you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The Swing Add On's package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with the Swing Add On's package.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.keang.swingaddons;

import java.awt.Component;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowFocusListener;
import java.util.concurrent.Callable;
import javax.swing.Timer;

/**
 * A mouse listener to determine if a mouse is hovering over a component and if so
 * to perform an action such as popping a dialog. 
 * <br>
 * Hovering is defined as the mouse entering a component and staying still
 * for a predetermined amount of time. Once the hover has been detected this detector will not
 * trigger again until the mouse has exited from the component and re-entered it.
 * <br>
 * This instance will register itself with the appropriate mouse listeners on the owning component. 
 * To de-register this hover detector from the owning component call the remove() method. Once removed
 * this hover detector can't be used again.
 *<br>
 * The action taken when a hover is detected is defined in the Callable object. If a window is
 * displayed then returning a reference to the window from the Callable object will allow this
 * hover detector to spot when this window no longer has focus and to close the window and
 * reset the hover detector appropriately. If this functionality is not required return null, if
 * an alternative functionality is required override the handleObject(Object obj) method.
 *
 *
 * @author A.G.Docherty
 * @date 22 Jan 2010
 *
 */
public class HoverDetector extends MouseAdapter
{
private static final int TIME_BEFORE_TRIGGER = 500;
private boolean triggered = false;
private Component owner;
private Timer t;
private Callable<Component> command;
private int triggerTime;

/**
 * Creates a hover detector and registers it with component. The trigger
 * delay time is the constant TIME_BEFORE_TRIGGER.
 * 
 * @param parent - the component to register for hover detection
 * @param r - the command to run on hover detection
 */
public HoverDetector(Component parent, Callable<Component> r)
    {
    this(parent, r, TIME_BEFORE_TRIGGER);
    }

/**
 * Creates a hover detector and registers it with the component.
 * 
 * @param parent - the component to register for hover detection
 * @param r - the command to run on hover detection
 * @param time - the delay time before triggering a hover
 */
public HoverDetector(Component parent, Callable<Component> r, int time)
    {
    owner = parent;
    command = r;
    triggerTime = time;
    
    owner.addMouseListener(this);
    owner.addMouseMotionListener(this);

    // create a timeout to wait for the mouse to hover for 'n' millis
    t = new Timer(triggerTime, new ActionListener()
        {
        @Override
        public void actionPerformed(ActionEvent e)
            {
            try
                {
                Component obj = command.call();

                if (obj != null )
                    {
                    handleObject(obj);
                    }
                }
            catch ( Exception e1 )
                {
                // the call failed - there is nothing we can do other than
                // print out the exception
                e1.printStackTrace();
                }
            }
        });

    t.setRepeats(false);
    }

/**
 * This method is called if an object is returned from the Callable object.
 * <br>
 * This implementation only handles Window type objects.
 * <br> 
 * Adds a window focus listener to the Window to listen for lost focus events.
 * The window will be closed if it loses focus and this hover detector will be reset if the
 * mouse is not currently over the owning component.
 * <br>
 * Override this method to provide different behaviour.
 * 
 * @param obj - the object returned from the Callable object
 */
protected void handleObject(Component obj)
    {
    if ( !(obj instanceof Window) )
        return;
    
    final Window w = (Window)obj;
    
    w.addWindowFocusListener(new WindowFocusListener() 
        {
        @Override
        public void windowLostFocus(WindowEvent e)
            {
            // close the window
            w.setVisible(false);
            w.dispose();
            
            // if the mouse is not over the component force a reset
            if ( owner.isVisible() && owner.getMousePosition() == null )
                reset();
            }
        
        @Override
        public void windowGainedFocus(WindowEvent e)
            {
            // do nothing
            }
        });
    }

/**
 * De-registers the hover detector from the component. Once called the
 * detector can not be used again.
 */
public void remove()
    {
    owner.removeMouseListener(this);
    owner.removeMouseMotionListener(this);
    }

/**
 * Forcibly resets the trigger mechanism. 
 * 
 * This is required if the trigger causes some without generating a mouseExited
 * event.
 */
public void reset()
    {
    triggered = false;
    }

@Override
public void mouseEntered(MouseEvent e)
    {
    // if already triggered ignore this event
    if ( triggered )
        return;

    triggered = true;

    t.start();
    }

@Override
public void mouseMoved(MouseEvent e)
    {
    if ( t.isRunning() )
        {
        // mouse has moved so restart the hover timeout
        t.restart();
        }
    }

@Override
public void mouseExited(MouseEvent e)
    {
    if ( t.isRunning() )
        {
        // the mouse has left the component so cancel the timeout
        t.stop();
        }

    // The check to see if we are no longer in the component even though mouseExited has been called
    // is to prevent repeated triggers if the trigger caused a dialog to display and the dialog
    // closes whilst the mouse is still over the component
    if ( !owner.contains(e.getPoint()) )
        triggered = false;
    }

}
