/** 
 * Copyright (c) 2010 Keang Ltd
 * 
 * This file is part of the Swing Add On's package.
 *
 * The Swing Add On's package is free software: you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The Swing Add On's package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with the Swing Add On's package.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.keang.swingaddons;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Insets;
import java.util.EnumSet;

import javax.swing.border.LineBorder;

/**
 * A Border that draws a single line along the specified edges of the component.
 * If drawing lines along all 4 edges use the javax.swing.border.LineBorder class as it
 * is more efficient.
 *
 * Combinations of line thicknesses can be obtained by using the
 * java.swing.border.CompoundBorder class to combine multiple borders ie
 * To draw thin line along the top edge and a thicker line along the bottom edge use:
 * <pre>
 * setBorder(new CompoundBorder(new SingleLineBorder(Side.TOP, Color.BLACK, 1),
 *           new SingleLineBorder(Side.BOTTOM, Color.BLACK, 2)));
 * </pre>
 * @author A.G.Docherty
 * @date 16 Jan 2010
 *
 */
public class SelectableLineBorder extends LineBorder
{
private static final long serialVersionUID = -7735274085177253488L;

/**
 * The sides of the component
 *
 * @author A.G.Docherty
 * @date 16 Jan 2010
 *
 */
public enum Side { LEFT, RIGHT, TOP, BOTTOM; }

private final EnumSet<Side> sides;

/**
 * Creates a line border on one side of the component with the specified color
 * and thickness of 1.
 *
 * @param side the side to draw one
 * @param color the color of the border
 */
public SelectableLineBorder(Side side, Color color)
    {
    this(side, color, 1);
    }

/**
 * Creates a line border on one side of the component with the specified color and thickness.
 *
 * @param side the side to draw one
 * @param color the color of the border
 * @param width the thickness of the border
 */
public SelectableLineBorder(Side side, Color color, int width)
    {
    super(color, width);

    this.sides = EnumSet.of(side);
    }

/**
 * Creates a line border on the specified sides of the component with the specified color and thickness.
 *
 * @param side the sides to draw one
 * @param color the color of the border
 * @param width the thickness of the border
 */
public SelectableLineBorder(EnumSet<Side> side, Color color, int width)
    {
    super(color, width);

    this.sides = side;
    }

/**
 * Paints the border for the specified component with the
 * specified position and size.
 * @param c the component for which this border is being painted
 * @param g the paint graphics
 * @param x the x position of the painted border
 * @param y the y position of the painted border
 * @param width the width of the painted border
 * @param height the height of the painted border
 */
@Override
public void paintBorder(Component c, Graphics g, int x, int y, int width, int height)
    {
    Color oldColor = g.getColor();
    g.setColor(lineColor);
    g.translate(x, y);
    
    for ( Side side : sides )
        {
        for ( int i = 0; i < thickness; i++ )
            {
            switch ( side )
                {
                case LEFT:
                    g.drawLine(i, 0, i, height);
                    break;
                case RIGHT:
                	int x1 = width-i-1;
                    g.drawLine(x1, 0, x1, height);
                    break;
                case TOP:
                    g.drawLine(0, i, width, i);
                    break;
                case BOTTOM:
                	int y1 = height-i-1;
                    g.drawLine(0, y1, width, y1);
                    break;
                default:
                    break;
                }
            }
        }

    g.setColor(oldColor);
    }

/**
 * Returns the insets of the border.
 * @param c the component for which this border insets value applies
 */
@Override
public Insets getBorderInsets(Component c)
    {
    return new Insets(sides.contains(Side.TOP)?thickness:0,
                      sides.contains(Side.LEFT)?thickness:0,
                      sides.contains(Side.BOTTOM)?thickness:0,
                      sides.contains(Side.RIGHT)?thickness:0);
    }

/**
 * Re-initialise the insets parameter with this Border's current Insets.
 * @param c the component for which this border insets value applies
 * @param insets the object to be re-initialised
 */
@Override
public Insets getBorderInsets(Component c, Insets insets)
    {
    insets.left = sides.contains(Side.LEFT)?thickness:0;
    insets.top = sides.contains(Side.TOP)?thickness:0;
    insets.right = sides.contains(Side.RIGHT)?thickness:0;
    insets.bottom = sides.contains(Side.BOTTOM)?thickness:0;

    return insets;
    }


}
