/** 
 * Copyright (c) 2010 Keang Ltd.
 * 
 * This file is part of the Swing Add On's package.
 *
 * The Swing Add On's package is free software: you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as published 
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * The Swing Add On's package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with the Swing Add On's package.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.keang.swingaddons;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager;

/**
 * The <code>VariableGridLayout</code> class is a layout manager that
 * lays out a container's components in a rectangular grid.
 * The container is divided into rectangles where the width of a
 * column is a ratio of the preferred width of the widest component in
 * each column and the height of a row is a ratio of the preferred
 * height of the highest component in each row.
 * One component is placed in each rectangle.
 * <p>
 * The rectangles are proportionally sized to fill the panel, to prevent 
 * the rectangles from being stretched to fill the panel call the 
 * <tt>setFixedSize(boolean, boolean)</tt> method. By default the layout 
 * manager will stretch the components to fill the panel. 
 * <p>
 * When both the number of rows and the number of columns have
 * been set to non-zero values, either by a constructor or
 * by the <tt>setRows</tt> and <tt>setColumns</tt> methods, the number of
 * columns specified is ignored.  Instead, the number of
 * columns is determined from the specified number of rows
 * and the total number of components in the layout. So, for
 * example, if three rows and two columns have been specified
 * and nine components are added to the layout, they will
 * be displayed as three rows of three columns.  Specifying
 * the number of columns affects the layout only when the
 * number of rows is set to zero.
 * 
 * This class is based on the original java.awt.GridLayout code by Arthur van Hoff
 * version 1.42, 11/17/05
 */
public class VariableGridLayout implements LayoutManager, java.io.Serializable {
    /*
     * serialVersionUID
     */
    private static final long serialVersionUID = -7804673224730901L;

    /**
     * This is the horizontal gap (in pixels) which specifies the space
     * between columns.  They can be changed at any time.
     * This should be a non-negative integer.
     *
     * @serial
     * @see #getHgap()
     * @see #setHgap(int)
     */
    int hgap;
    /**
     * This is the vertical gap (in pixels) which specifies the space
     * between rows.  They can be changed at any time.
     * This should be a non negative integer.
     *
     * @serial
     * @see #getVgap()
     * @see #setVgap(int)
     */
    int vgap;
    /**
     * This is the number of rows specified for the grid.  The number
     * of rows can be changed at any time.
     * This should be a non negative integer, where '0' means
     * 'any number' meaning that the number of Rows in that
     * dimension depends on the other dimension.
     *
     * @serial
     * @see #getRows()
     * @see #setRows(int)
     */
    int rows;
    /**
     * This is the number of columns specified for the grid.  The number
     * of columns can be changed at any time.
     * This should be a non negative integer, where '0' means
     * 'any number' meaning that the number of Columns in that
     * dimension depends on the other dimension.
     *
     * @serial
     * @see #getColumns()
     * @see #setColumns(int)
     */
    int cols;

    /**
     * Fixes the horizontal axis to prevent the columns being stretched to fill the panel.
     * By default this is false.
     * 
     * @see #setFixedSize(boolean, boolean)
     */
    boolean fixH = false;

    /**
     * Fixes the vertical axis to prevent the rows being stretched to fill the panel.
     * By default this is false.
     * 
     * @see #setFixedSize(boolean, boolean)
     */
    boolean fixV = false;

    /**
     * Creates a grid layout with a default of one column per component,
     * in a single row.
     */
    public VariableGridLayout() {
	this(1, 0, 0, 0);
    }

    /**
     * Creates a grid layout with the specified number of rows and
     * columns.
     * <p>
     * One, but not both, of <code>rows</code> and <code>cols</code> can
     * be zero, which means that any number of objects can be placed in a
     * row or in a column.
     * @param     r   the rows, with the value zero meaning
     *                   any number of rows.
     * @param     c   the columns, with the value zero meaning
     *                   any number of columns.
     */
    public VariableGridLayout(int r, int c) {
	this(r, c, 0, 0);
    }

    /**
     * Creates a grid layout with the specified number of rows and
     * columns.
     * <p>
     * In addition, the horizontal and vertical gaps are set to the
     * specified values. Horizontal gaps are placed between each
     * of the columns. Vertical gaps are placed between each of
     * the rows.
     * <p>
     * One, but not both, of <code>rows</code> and <code>cols</code> can
     * be zero, which means that any number of objects can be placed in a
     * row or in a column.
     * <p>
     * All <code>VariableGridLayout</code> constructors defer to this one.
     * @param     r   the rows, with the value zero meaning
     *                   any number of rows
     * @param     c   the columns, with the value zero meaning
     *                   any number of columns
     * @param     hgp   the horizontal gap
     * @param     vgp   the vertical gap
     * @exception   IllegalArgumentException  if the value of both
     *			<code>rows</code> and <code>cols</code> is
     *			set to zero
     */
    public VariableGridLayout(int r, int c, int hgp, int vgp) {
	if ((r == 0) && (c == 0)) {
	    throw new IllegalArgumentException("rows and cols cannot both be zero");
	}
	this.rows = r;
	this.cols = c;
	this.hgap = hgp;
	this.vgap = vgp;
    }

    /**
     * Sets the can stretch state of the horizontal and vertical axis.
     * By default both the horizontal and vertical axes will be stretched.
     * Fixing the size of either or both axes will not stop the axis from 
     * being shrunk.
     *  
     * @param v - true to fix the vertical axis to prevent the rows being stretched to fill the panel.
     * @param h - true to fix the horizontal axis to prevent the columns being stretched to fill the panel.
     */
    public void setFixedSize(boolean v, boolean h) {
    fixV = v;
    fixH = h;
    }
    
    /**
     * Gets the number of rows in this layout.
     * @return    the number of rows in this layout
     */
    public int getRows() {
	return rows;
    }

    /**
     * Sets the number of rows in this layout to the specified value.
     * @param        r   the number of rows in this layout
     * @exception    IllegalArgumentException  if the value of both
     *               <code>rows</code> and <code>cols</code> is set to zero
     */
    public void setRows(int r) {
	if ((r == 0) && (this.cols == 0)) {
	    throw new IllegalArgumentException("rows and cols cannot both be zero");
	}
	this.rows = r;
    }

    /**
     * Gets the number of columns in this layout.
     * @return     the number of columns in this layout
     */
    public int getColumns() {
	return cols;
    }

    /**
     * Sets the number of columns in this layout to the specified value.
     * Setting the number of columns has no affect on the layout
     * if the number of rows specified by a constructor or by
     * the <tt>setRows</tt> method is non-zero. In that case, the number
     * of columns displayed in the layout is determined by the total
     * number of components and the number of rows specified.
     * @param        c   the number of columns in this layout
     * @exception    IllegalArgumentException  if the value of both
     *               <code>rows</code> and <code>cols</code> is set to zero
     */
    public void setColumns(int c) {
	if ((c == 0) && (this.rows == 0)) {
	    throw new IllegalArgumentException("rows and cols cannot both be zero");
	}
	this.cols = c;
    }

    /**
     * Gets the horizontal gap between components.
     * @return       the horizontal gap between components
     */
    public int getHgap() {
	return hgap;
    }

    /**
     * Sets the horizontal gap between components to the specified value.
     * @param        hgp   the horizontal gap between components
     */
    public void setHgap(int hgp) {
	this.hgap = hgp;
    }

    /**
     * Gets the vertical gap between components.
     * @return       the vertical gap between components
     */
    public int getVgap() {
	return vgap;
    }

    /**
     * Sets the vertical gap between components to the specified value.
     * @param         vgp  the vertical gap between components
     */
    public void setVgap(int vgp) {
	this.vgap = vgp;
    }

    /**
     * Adds the specified component with the specified name to the layout.
     * @param name the name of the component
     * @param comp the component to be added
     */
    @Override
    public void addLayoutComponent(String name, Component comp)
        {
        // do nothing
        }

    /**
     * Removes the specified component from the layout.
     * @param comp the component to be removed
     */
    @Override
    public void removeLayoutComponent(Component comp)
        {
        // do nothing
        }

    /**
     * Determines the preferred size of the container argument using
     * this grid layout.
     * <p>
     * The preferred width of a grid layout is the sum of the components
     * largest preferred width of each column
     * plus the horizontal padding times the number of columns
     * minus one, plus the left and right insets of the target container.
     * <p>
     * The preferred height of a grid layout is the sum of the components
     * largest preferred height of each row times the number of
     * rows, plus the vertical padding times the number of rows minus one,
     * plus the top and bottom insets of the target container.
     *
     * @param     parent   the container in which to do the layout
     * @return    the preferred dimensions to lay out the
     *                      subcomponents of the specified container
     */
    @Override
    public Dimension preferredLayoutSize(Container parent) {
      synchronized (parent.getTreeLock()) {
	Insets insets = parent.getInsets();
	int ncomponents = parent.getComponentCount();
	int nrows = rows;
	int ncols = cols;

	if (nrows > 0) {
	    ncols = (ncomponents + nrows - 1) / nrows;
	} else {
	    nrows = (ncomponents + ncols - 1) / ncols;
	}

    int[] colWidths = new int[ncols];
    int[] rowHeights = new int[nrows];
    Component[] comps = parent.getComponents();

	for (int i = 0 ; i < nrows ; i++) {
        for ( int j = 0; j < ncols; ++j ) {
            int idx = i * ncols + j;

            if ( idx >= comps.length ) {
                break;
                }

            Component comp = comps[idx];
    	    Dimension d = comp.getPreferredSize();

    	    if (colWidths[j] < d.width) {
        		colWidths[j] = d.width;
            }

    	    if (rowHeights[i] < d.height) {
    		    rowHeights[i] = d.height;
            }
        }
	}

    int w = 0;
	for (int i = 0 ; i < ncols ; i++) {
        w += colWidths[i];
    }

    int h = 0;
	for (int i = 0 ; i < nrows ; i++) {
        h += rowHeights[i];
    }

	return new Dimension(insets.left + insets.right + w + (ncols-1)*hgap,
			     insets.top + insets.bottom + h + (nrows-1)*vgap);
      }
    }

    /**
     * Determines the minimum size of the container argument using this
     * grid layout.
     * <p>
     * The minimum width of a grid layout is the sum of the components
     * largest minimum width of each column
     * plus the horizontal padding times the number of columns
     * minus one, plus the left and right insets of the target container.
     * <p>
     * The minimum height of a grid layout is the sum of the components
     * largest minimum height of each row times the number of
     * rows, plus the vertical padding times the number of rows minus one,
     * plus the top and bottom insets of the target container.
     *
     * @param       parent   the container in which to do the layout
     * @return      the minimum dimensions needed to lay out the
     *                      subcomponents of the specified container
     */
    @Override
    public Dimension minimumLayoutSize(Container parent) {
      synchronized (parent.getTreeLock()) {
	Insets insets = parent.getInsets();
	int ncomponents = parent.getComponentCount();
	int nrows = rows;
	int ncols = cols;

	if (nrows > 0) {
	    ncols = (ncomponents + nrows - 1) / nrows;
	} else {
	    nrows = (ncomponents + ncols - 1) / ncols;
	}

    int[] colWidths = new int[ncols];
    int[] rowHeights = new int[nrows];
    Component[] comps = parent.getComponents();

	for (int i = 0 ; i < nrows ; i++) {
        for ( int j = 0; j < ncols; ++j ) {
            int idx = i * ncols + j;

            if ( idx >= comps.length ) {
                break;
                }

            Component comp = comps[idx];
    	    Dimension d = comp.getMinimumSize();

    	    if (colWidths[j] < d.width) {
        		colWidths[j] = d.width;
            }

    	    if (rowHeights[i] < d.height) {
    		    rowHeights[i] = d.height;
            }
        }
	}

    int w = 0;
	for (int i = 0 ; i < ncols ; i++) {
        w += colWidths[i];
    }

    int h = 0;
	for (int i = 0 ; i < nrows ; i++) {
        h += rowHeights[i];
    }

	return new Dimension(insets.left + insets.right + w + (ncols-1)*hgap,
			     insets.top + insets.bottom + h + (nrows-1)*vgap);
      }
    }

    /**
     * Lays out the specified container using this layout.
     * <p>
     * This method reshapes the components in the specified target
     * container in order to satisfy the constraints of the
     * <code>VariableGridLayout</code> object.
     * <p>
     * The variable grid layout manager determines the size of individual
     * components by dividing the free space in the container into
     * unequal portions according to relative maximum preferred sizes of
     * largest component in each row and column in the layout.
     * The container's free space equals the container's
     * size minus any insets and any specified horizontal or vertical
     * gap.
     *
     * @param      parent   the container in which to do the layout
     * @see        java.awt.Container
     * @see        java.awt.Container#doLayout
     */
    @Override
    public void layoutContainer(Container parent) {
      synchronized (parent.getTreeLock()) {
	Insets insets = parent.getInsets();
	int ncomponents = parent.getComponentCount();
	int nrows = rows;
	int ncols = cols;
	boolean ltr = parent.getComponentOrientation().isLeftToRight();

	if (ncomponents == 0) {
	    return;
	}
	if (nrows > 0) {
	    ncols = (ncomponents + nrows - 1) / nrows;
	} else {
	    nrows = (ncomponents + ncols - 1) / ncols;
	}
	int w = parent.getWidth() - (insets.left + insets.right)  - (ncols - 1) * hgap;
	int h = parent.getHeight() - (insets.top + insets.bottom) - (nrows - 1) * vgap;

    int[] colWidths = new int[ncols];
    int[] rowHeights = new int[nrows];
    Component[] comps = parent.getComponents();

    // find the largest width and height for each column and row respectively
	for (int i = 0 ; i < nrows ; i++) {
        for ( int j = 0; j < ncols; ++j ) {
            int idx = i * ncols + j;

            if ( idx >= comps.length ) {
                break;
                }

            Component comp = comps[idx];
    	    Dimension d = comp.getPreferredSize();

    	    if (colWidths[j] < d.width) {
        		colWidths[j] = d.width;
            }

    	    if (rowHeights[i] < d.height) {
    		    rowHeights[i] = d.height;
            }
        }
	}

    int prefW = 0;
	for (int i = 0 ; i < ncols ; i++) {
        prefW += colWidths[i];
    }

    int prefH = 0;
	for (int i = 0 ; i < nrows ; i++) {
        prefH += rowHeights[i];
    }

    double ratioW = 1.0;
    
    // if the ratio is less than 1.0 or we can stretch
    if ( w < prefW || !fixH )
        ratioW = (double)w/prefW;
    
    double ratioH = 1.0;
    
    // if the ratio is less than 1.0 or we can stretch
    if ( h < prefH || !fixV )
        ratioH = (double)h/prefH;

	if (ltr) {
	    // left to right orientation
	    for (int c = 0, x = insets.left ; c < ncols ; c++, x += w + hgap) {
	    	w = (int)(colWidths[c] * ratioW);
			for (int r = 0, y = insets.top ; r < nrows ; r++, y += h + vgap) {
			    int i = r * ncols + c;
			    if (i < ncomponents) {
	    			Component comp = parent.getComponent(i);
	
	                h = (int)(rowHeights[r] * ratioH);
	
	    			comp.setBounds(x, y, w, h);
			    }
			}
	    }
	} else {
	    // right to left orientation
	    for (int c = 0, x = parent.getWidth() - insets.right - w; c < ncols ; c++, x -= w + hgap) {
	        w = (int)(colWidths[c] * ratioW);
			for (int r = 0, y = insets.top ; r < nrows ; r++, y += h + vgap) {
			    int i = r * ncols + c;
			    if (i < ncomponents) {
	    			Component comp = parent.getComponent(i);
	
	                h = (int)(rowHeights[r] * ratioH);
	
	    			comp.setBounds(x, y, w, h);
			    }
			}
	    }
	}
      }
    }

    /**
     * Returns the string representation of this grid layout's values.
     * @return     a string representation of this grid layout
     */
    @Override
    public String toString() 
        {
        return getClass().getName() + "[hgap=" + hgap + ",vgap=" + vgap +
	    			       ",rows=" + rows + ",cols=" + cols + "]";
        }
}
